"""Generated implementation of feature_store."""

# WARNING DO NOT EDIT
# This code was generated from feature-store.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..attribute import Attribute
from ..cluster import ClusterId
from ..commit import CommitId
from ..destination_reference import DestinationReference
from ..feature_set import FeatureSetId
from ..label import Label
from ..schedule import Schedule


@dataclasses.dataclass(frozen=True)
class FeatureStoreId:
    """Unique identifier for a feature store.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreId:
        """Validate and parse JSON data into an instance of FeatureStoreId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing FeatureStoreId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> FeatureStoreId:
        """Parse a JSON string such as a dictionary key."""
        return FeatureStoreId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class FeatureStoreName:
    """Unique name for a feature store.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreName:
        """Validate and parse JSON data into an instance of FeatureStoreName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing FeatureStoreName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> FeatureStoreName:
        """Parse a JSON string such as a dictionary key."""
        return FeatureStoreName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class FeatureStoreVersionId:
    """Unique identifier of a specific version of a feature store.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreVersionId:
        """Validate and parse JSON data into an instance of FeatureStoreVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing FeatureStoreVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> FeatureStoreVersionId:
        """Parse a JSON string such as a dictionary key."""
        return FeatureStoreVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class FeatureStore:
    """Definition of a feature store.
    
    Args:
        id (FeatureStoreId): A data field.
        name (FeatureStoreName): A data field.
        description (str): A data field.
        labels (typing.List[Label]): A data field.
        attributes (typing.List[Attribute]): A data field.
        featureSet (FeatureSetId): A data field.
        enabled (bool): A data field.
        destinations (typing.List[DestinationReference]): A data field.
        cluster (ClusterId): A data field.
        schedule (Schedule): A data field.
        startDate (typing.Optional[datetime.date]): A data field.
        endDate (typing.Optional[datetime.date]): A data field.
        runDateOffset (typing.Optional[int]): A data field.
        versionTarget (typing.Optional[VersionTarget]): A data field.
        version (FeatureStoreVersionId): A data field.
    """
    
    id: FeatureStoreId
    name: FeatureStoreName
    description: str
    labels: typing.List[Label]
    attributes: typing.List[Attribute]
    featureSet: FeatureSetId
    enabled: bool
    destinations: typing.List[DestinationReference]
    cluster: ClusterId
    schedule: Schedule
    startDate: typing.Optional[datetime.date]
    endDate: typing.Optional[datetime.date]
    runDateOffset: typing.Optional[int]
    versionTarget: typing.Optional[VersionTarget]
    version: FeatureStoreVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStore data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": FeatureStoreId.json_schema(),
                "name": FeatureStoreName.json_schema(),
                "description": {
                    "type": "string"
                },
                "labels": {
                    "type": "array",
                    "item": Label.json_schema()
                },
                "attributes": {
                    "type": "array",
                    "item": Attribute.json_schema()
                },
                "featureSet": FeatureSetId.json_schema(),
                "enabled": {
                    "type": "boolean"
                },
                "destinations": {
                    "type": "array",
                    "item": DestinationReference.json_schema()
                },
                "cluster": ClusterId.json_schema(),
                "schedule": Schedule.json_schema(),
                "startDate": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date"},
                    ]
                },
                "endDate": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date"},
                    ]
                },
                "runDateOffset": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "versionTarget": {
                    "oneOf": [
                        {"type": "null"},
                        VersionTarget.json_schema(),
                    ]
                },
                "version": FeatureStoreVersionId.json_schema()
            },
            "required": [
                "id",
                "name",
                "description",
                "labels",
                "attributes",
                "featureSet",
                "enabled",
                "destinations",
                "cluster",
                "schedule",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStore:
        """Validate and parse JSON data into an instance of FeatureStore.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStore.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStore(
                id=FeatureStoreId.from_json(data["id"]),
                name=FeatureStoreName.from_json(data["name"]),
                description=str(data["description"]),
                labels=[Label.from_json(v) for v in data["labels"]],
                attributes=[Attribute.from_json(v) for v in data["attributes"]],
                featureSet=FeatureSetId.from_json(data["featureSet"]),
                enabled=bool(data["enabled"]),
                destinations=[DestinationReference.from_json(v) for v in data["destinations"]],
                cluster=ClusterId.from_json(data["cluster"]),
                schedule=Schedule.from_json(data["schedule"]),
                startDate=(
                    lambda v: v and datetime.date.fromisoformat(v)
                )(
                    data.get("startDate", None)
                ),
                endDate=(
                    lambda v: v and datetime.date.fromisoformat(v)
                )(
                    data.get("endDate", None)
                ),
                runDateOffset=(lambda v: v and int(v))(data.get("runDateOffset", None)),
                versionTarget=(
                    lambda v: v and VersionTarget.from_json(v)
                )(
                    data.get("versionTarget", None)
                ),
                version=FeatureStoreVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureStore",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "labels": [v.to_json() for v in self.labels],
            "attributes": [v.to_json() for v in self.attributes],
            "featureSet": self.featureSet.to_json(),
            "enabled": self.enabled,
            "destinations": [v.to_json() for v in self.destinations],
            "cluster": self.cluster.to_json(),
            "schedule": self.schedule.to_json(),
            "startDate": (lambda v: v and v.isoformat())(self.startDate),
            "endDate": (lambda v: v and v.isoformat())(self.endDate),
            "runDateOffset": (lambda v: v and int(v))(self.runDateOffset),
            "versionTarget": (lambda v: v and v.to_json())(self.versionTarget),
            "version": self.version.to_json()
        }


@dataclasses.dataclass(frozen=True)
class VersionTarget(abc.ABC):
    """Code version to execute when running a feature store."""
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> VersionTarget:
        """JSON schema for variant VersionTarget.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> VersionTarget:
        """Validate and parse JSON data into an instance of VersionTarget.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of VersionTarget.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing VersionTarget", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class CommitTarget(VersionTarget):
    """Execute code from a specific commit.
    
    Args:
        commitId (CommitId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "committarget"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    commitId: CommitId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CommitTarget data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "commitId": CommitId.json_schema()
            },
            "required": [
                "adt_type",
                "commitId",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CommitTarget:
        """Validate and parse JSON data into an instance of CommitTarget.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CommitTarget.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CommitTarget(
                commitId=CommitId.from_json(data["commitId"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CommitTarget",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "commitId": self.commitId.to_json()
        }


@dataclasses.dataclass(frozen=True)
class BranchTarget(VersionTarget):
    """Execute latest code from a branch.
    
    Args:
        branchName (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "branchtarget"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    branchName: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BranchTarget data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "branchName": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "branchName",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BranchTarget:
        """Validate and parse JSON data into an instance of BranchTarget.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BranchTarget.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BranchTarget(
                branchName=str(data["branchName"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BranchTarget",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "branchName": str(self.branchName)
        }



"""Generated implementation of entity_mapping."""

# WARNING DO NOT EDIT
# This code was generated from entity-mapping.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..entity import EntityId
from ..feature_id import FeatureId


@dataclasses.dataclass(frozen=True)
class EntityMappingId:
    """Unique identifier for entity mappings.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityMappingId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityMappingId:
        """Validate and parse JSON data into an instance of EntityMappingId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityMappingId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityMappingId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing EntityMappingId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> EntityMappingId:
        """Parse a JSON string such as a dictionary key."""
        return EntityMappingId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EntityMappingVersionId:
    """Unique identifier for versions of an entity mapping.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityMappingVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityMappingVersionId:
        """Validate and parse JSON data into an instance of EntityMappingVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityMappingVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityMappingVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing EntityMappingVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> EntityMappingVersionId:
        """Parse a JSON string such as a dictionary key."""
        return EntityMappingVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class EntityMapping:
    """A mapping defined between between two entities.
    
    Args:
        id (EntityMappingId): A data field.
        from_ (EntityId): A data field.
        to (EntityId): A data field.
        mapping (FeatureId): A data field.
        version (EntityMappingVersionId): A data field.
    """
    
    id: EntityMappingId
    from_: EntityId
    to: EntityId
    mapping: FeatureId
    version: EntityMappingVersionId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EntityMapping data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": EntityMappingId.json_schema(),
                "from": EntityId.json_schema(),
                "to": EntityId.json_schema(),
                "mapping": FeatureId.json_schema(),
                "version": EntityMappingVersionId.json_schema()
            },
            "required": [
                "id",
                "from",
                "to",
                "mapping",
                "version",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EntityMapping:
        """Validate and parse JSON data into an instance of EntityMapping.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EntityMapping.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EntityMapping(
                id=EntityMappingId.from_json(data["id"]),
                from_=EntityId.from_json(data["from"]),
                to=EntityId.from_json(data["to"]),
                mapping=FeatureId.from_json(data["mapping"]),
                version=EntityMappingVersionId.from_json(data["version"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EntityMapping",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "from": self.from_.to_json(),
            "to": self.to.to_json(),
            "mapping": self.mapping.to_json(),
            "version": self.version.to_json()
        }



"""Generated implementation of aggregate."""

# WARNING DO NOT EDIT
# This code was generated from aggregate.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401


class AggregateExpression(enum.Enum):
    """Aggregate expressions for Anaml Event features."""
    Sum = "sum"
    Count = "count"
    CountDistinct = "countdistinct"
    Avg = "avg"
    Std = "std"
    Max = "max"
    Min = "min"
    MaxBy = "maxby"
    MinBy = "minby"
    Last = "last"
    PercentageChange = "percentagechange"
    AbsoluteChange = "absolutechange"
    StandardScore = "standardscore"
    CollectList = "collectlist"
    CollectSet = "collectset"
    BasketSum = "basketsum"
    BasketLast = "basketlast"
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'AggregateExpression'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "sum",
                        "count",
                        "countdistinct",
                        "avg",
                        "std",
                        "max",
                        "min",
                        "maxby",
                        "minby",
                        "last",
                        "percentagechange",
                        "absolutechange",
                        "standardscore",
                        "collectlist",
                        "collectset",
                        "basketsum",
                        "basketlast",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AggregateExpression:
        """Validate and parse JSON data into an instance of AggregateExpression.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AggregateExpression.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AggregateExpression(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing AggregateExpression", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> AggregateExpression:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of AggregateExpression.
        """
        return AggregateExpression(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)



"""Generated implementation of summary_statistics."""

# WARNING DO NOT EDIT
# This code was generated from summary-statistics.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401


@dataclasses.dataclass(frozen=True)
class SummaryStatistics(abc.ABC):
    """Summary statistics for a feature.
    
    Args:
        countNulls (typing.Optional[int]): A data field.
        dataType (typing.Optional[str]): A data field.
        featureName (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    countNulls: typing.Optional[int]
    dataType: typing.Optional[str]
    featureName: str
    
    @classmethod
    def json_schema(cls) -> SummaryStatistics:
        """JSON schema for variant SummaryStatistics.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SummaryStatistics:
        """Validate and parse JSON data into an instance of SummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing SummaryStatistics", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class NumericalSummaryStatistics(SummaryStatistics):
    """Summary statistics for a numerical feature.
    
    Args:
        featureName (str): A data field.
        dataType (typing.Optional[str]): A data field.
        count (int): A data field.
        countNulls (typing.Optional[int]): A data field.
        min (typing.Optional[float]): A data field.
        max (typing.Optional[float]): A data field.
        mean (typing.Optional[float]): A data field.
        stdDev (typing.Optional[float]): A data field.
        quantiles (typing.List[float]): A data field.
        densities (typing.List[Density]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "numerical"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    featureName: str
    dataType: typing.Optional[str]
    count: int
    countNulls: typing.Optional[int]
    min: typing.Optional[float]
    max: typing.Optional[float]
    mean: typing.Optional[float]
    stdDev: typing.Optional[float]
    quantiles: typing.List[float]
    densities: typing.List[Density]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for NumericalSummaryStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "featureName": {
                    "type": "string"
                },
                "dataType": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "count": {
                    "type": "integer"
                },
                "countNulls": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "min": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "number"},
                    ]
                },
                "max": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "number"},
                    ]
                },
                "mean": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "number"},
                    ]
                },
                "stdDev": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "number"},
                    ]
                },
                "quantiles": {
                    "type": "array",
                    "item": {"type": "number"}
                },
                "densities": {
                    "type": "array",
                    "item": Density.json_schema()
                }
            },
            "required": [
                "adt_type",
                "featureName",
                "count",
                "quantiles",
                "densities",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> NumericalSummaryStatistics:
        """Validate and parse JSON data into an instance of NumericalSummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of NumericalSummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return NumericalSummaryStatistics(
                featureName=str(data["featureName"]),
                dataType=(lambda v: v and str(v))(data.get("dataType", None)),
                count=int(data["count"]),
                countNulls=(lambda v: v and int(v))(data.get("countNulls", None)),
                min=(lambda v: v and float(v))(data.get("min", None)),
                max=(lambda v: v and float(v))(data.get("max", None)),
                mean=(lambda v: v and float(v))(data.get("mean", None)),
                stdDev=(lambda v: v and float(v))(data.get("stdDev", None)),
                quantiles=[float(v) for v in data["quantiles"]],
                densities=[Density.from_json(v) for v in data["densities"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing NumericalSummaryStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "featureName": str(self.featureName),
            "dataType": (lambda v: v and str(v))(self.dataType),
            "count": self.count,
            "countNulls": (lambda v: v and v)(self.countNulls),
            "min": (lambda v: v and v)(self.min),
            "max": (lambda v: v and v)(self.max),
            "mean": (lambda v: v and v)(self.mean),
            "stdDev": (lambda v: v and v)(self.stdDev),
            "quantiles": [v for v in self.quantiles],
            "densities": [v.to_json() for v in self.densities]
        }


@dataclasses.dataclass(frozen=True)
class DateSummaryStatistics(SummaryStatistics):
    """Summary statistics for a date feature.
    
    Args:
        featureName (str): A data field.
        dataType (typing.Optional[str]): A data field.
        count (int): A data field.
        countNulls (typing.Optional[int]): A data field.
        min (typing.Optional[datetime.date]): A data field.
        max (typing.Optional[datetime.date]): A data field.
        quantiles (typing.List[datetime.date]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "date"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    featureName: str
    dataType: typing.Optional[str]
    count: int
    countNulls: typing.Optional[int]
    min: typing.Optional[datetime.date]
    max: typing.Optional[datetime.date]
    quantiles: typing.List[datetime.date]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for DateSummaryStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "featureName": {
                    "type": "string"
                },
                "dataType": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "count": {
                    "type": "integer"
                },
                "countNulls": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "min": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date"},
                    ]
                },
                "max": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date"},
                    ]
                },
                "quantiles": {
                    "type": "array",
                    "item": {"type": "string", "format": "date"}
                }
            },
            "required": [
                "adt_type",
                "featureName",
                "count",
                "quantiles",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> DateSummaryStatistics:
        """Validate and parse JSON data into an instance of DateSummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of DateSummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return DateSummaryStatistics(
                featureName=str(data["featureName"]),
                dataType=(lambda v: v and str(v))(data.get("dataType", None)),
                count=int(data["count"]),
                countNulls=(lambda v: v and int(v))(data.get("countNulls", None)),
                min=(lambda v: v and datetime.date.fromisoformat(v))(data.get("min", None)),
                max=(lambda v: v and datetime.date.fromisoformat(v))(data.get("max", None)),
                quantiles=[datetime.date.fromisoformat(v) for v in data["quantiles"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing DateSummaryStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "featureName": str(self.featureName),
            "dataType": (lambda v: v and str(v))(self.dataType),
            "count": self.count,
            "countNulls": (lambda v: v and v)(self.countNulls),
            "min": (lambda v: v and v.isoformat())(self.min),
            "max": (lambda v: v and v.isoformat())(self.max),
            "quantiles": [v.isoformat() for v in self.quantiles]
        }


@dataclasses.dataclass(frozen=True)
class CategoricalSummaryStatistics(SummaryStatistics):
    """Summary statistics for a categorical feature.
    
    Args:
        featureName (str): A data field.
        dataType (typing.Optional[str]): A data field.
        count (int): A data field.
        countNulls (typing.Optional[int]): A data field.
        categoryFrequencies (typing.Optional[typing.List[CategoryFrequency]]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "categorical"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    featureName: str
    dataType: typing.Optional[str]
    count: int
    countNulls: typing.Optional[int]
    categoryFrequencies: typing.Optional[typing.List[CategoryFrequency]]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CategoricalSummaryStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "featureName": {
                    "type": "string"
                },
                "dataType": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "count": {
                    "type": "integer"
                },
                "countNulls": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "categoryFrequencies": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": CategoryFrequency.json_schema()},
                    ]
                }
            },
            "required": [
                "adt_type",
                "featureName",
                "count",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CategoricalSummaryStatistics:
        """Validate and parse JSON data into an instance of CategoricalSummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CategoricalSummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CategoricalSummaryStatistics(
                featureName=str(data["featureName"]),
                dataType=(lambda v: v and str(v))(data.get("dataType", None)),
                count=int(data["count"]),
                countNulls=(lambda v: v and int(v))(data.get("countNulls", None)),
                categoryFrequencies=(
                    lambda v: v and [CategoryFrequency.from_json(v) for v in v]
                )(
                    data.get("categoryFrequencies", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CategoricalSummaryStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "featureName": str(self.featureName),
            "dataType": (lambda v: v and str(v))(self.dataType),
            "count": self.count,
            "countNulls": (lambda v: v and v)(self.countNulls),
            "categoryFrequencies": (lambda v: v and [v.to_json() for v in v])(self.categoryFrequencies)
        }


@dataclasses.dataclass(frozen=True)
class DefaultSummaryStatistics(SummaryStatistics):
    """Summary statistics for an arbitrary feature.
    
    Args:
        featureName (str): A data field.
        dataType (typing.Optional[str]): A data field.
        count (int): A data field.
        countNulls (typing.Optional[int]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "default"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    featureName: str
    dataType: typing.Optional[str]
    count: int
    countNulls: typing.Optional[int]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for DefaultSummaryStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "featureName": {
                    "type": "string"
                },
                "dataType": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "count": {
                    "type": "integer"
                },
                "countNulls": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                }
            },
            "required": [
                "adt_type",
                "featureName",
                "count",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> DefaultSummaryStatistics:
        """Validate and parse JSON data into an instance of DefaultSummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of DefaultSummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return DefaultSummaryStatistics(
                featureName=str(data["featureName"]),
                dataType=(lambda v: v and str(v))(data.get("dataType", None)),
                count=int(data["count"]),
                countNulls=(lambda v: v and int(v))(data.get("countNulls", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing DefaultSummaryStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "featureName": str(self.featureName),
            "dataType": (lambda v: v and str(v))(self.dataType),
            "count": self.count,
            "countNulls": (lambda v: v and v)(self.countNulls)
        }


@dataclasses.dataclass(frozen=True)
class EmptySummaryStatistics(SummaryStatistics):
    """Summary statistics for an empty feature.
    
    Args:
        featureName (str): A data field.
        dataType (typing.Optional[str]): A data field.
        countNulls (typing.Optional[int]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "empty"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    featureName: str
    dataType: typing.Optional[str]
    countNulls: typing.Optional[int]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EmptySummaryStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "featureName": {
                    "type": "string"
                },
                "dataType": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "countNulls": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                }
            },
            "required": [
                "adt_type",
                "featureName",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EmptySummaryStatistics:
        """Validate and parse JSON data into an instance of EmptySummaryStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EmptySummaryStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EmptySummaryStatistics(
                featureName=str(data["featureName"]),
                dataType=(lambda v: v and str(v))(data.get("dataType", None)),
                countNulls=(lambda v: v and int(v))(data.get("countNulls", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EmptySummaryStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "featureName": str(self.featureName),
            "dataType": (lambda v: v and str(v))(self.dataType),
            "countNulls": (lambda v: v and v)(self.countNulls)
        }


@dataclasses.dataclass(frozen=True)
class Density:
    """Probability density at a specific value.
    
    Args:
        value (float): A data field.
        density (float): A data field.
    """
    
    value: float
    density: float
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Density data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "value": {
                    "type": "number"
                },
                "density": {
                    "type": "number"
                }
            },
            "required": [
                "value",
                "density",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Density:
        """Validate and parse JSON data into an instance of Density.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Density.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Density(
                value=float(data["value"]),
                density=float(data["density"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Density",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "value": self.value,
            "density": self.density
        }


@dataclasses.dataclass(frozen=True)
class CategoryFrequency:
    """Frequency of an individual categorical value.
    
    Args:
        category (str): A data field.
        frequency (int): A data field.
    """
    
    category: str
    frequency: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CategoryFrequency data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "category": {
                    "type": "string"
                },
                "frequency": {
                    "type": "integer"
                }
            },
            "required": [
                "category",
                "frequency",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CategoryFrequency:
        """Validate and parse JSON data into an instance of CategoryFrequency.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CategoryFrequency.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CategoryFrequency(
                category=str(data["category"]),
                frequency=int(data["frequency"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CategoryFrequency",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "category": str(self.category),
            "frequency": self.frequency
        }



#
#  Copyright 2021 Simple Machines Pty Ltd - All Rights Reserved.
#
#  This file is part of Anaml.
#
#  Unauthorized copying and/or distribution of this file, via any medium
#  is strictly prohibited.
#
"""Version metadata for the package."""

__version__ = "0.14.0"



"""Generated implementation of feature_store_run."""

# WARNING DO NOT EDIT
# This code was generated from feature-store-run.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..commit import CommitId
from ..feature_id import FeatureId
from ..feature_store import FeatureStoreId, FeatureStoreVersionId
from ..job_metrics import JobMetrics
from ..jobs import FeatureStoreRunId, RunStatus
from ..schedule import ScheduleState
from ..summary_statistics import SummaryStatistics


@dataclasses.dataclass(frozen=True)
class FeatureStoreRun:
    """Details of a feature store run.
    
    Args:
        id (FeatureStoreRunId): A data field.
        created (datetime.datetime): A data field.
        featureStoreId (FeatureStoreId): A data field.
        featureStoreVersionId (FeatureStoreVersionId): A data field.
        commitId (CommitId): A data field.
        runStartDate (datetime.date): A data field.
        runEndDate (datetime.date): A data field.
        status (RunStatus): A data field.
        errorMessage (typing.Optional[str]): A data field.
        scheduleState (typing.Optional[ScheduleState]): A data field.
        statistics (typing.Optional[FeatureStoreExecutionStatistics]): A data field.
    """
    
    id: FeatureStoreRunId
    created: datetime.datetime
    featureStoreId: FeatureStoreId
    featureStoreVersionId: FeatureStoreVersionId
    commitId: CommitId
    runStartDate: datetime.date
    runEndDate: datetime.date
    status: RunStatus
    errorMessage: typing.Optional[str]
    scheduleState: typing.Optional[ScheduleState]
    statistics: typing.Optional[FeatureStoreExecutionStatistics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreRun data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": FeatureStoreRunId.json_schema(),
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "featureStoreId": FeatureStoreId.json_schema(),
                "featureStoreVersionId": FeatureStoreVersionId.json_schema(),
                "commitId": CommitId.json_schema(),
                "runStartDate": {
                    "type": "string",
                    "format": "date"
                },
                "runEndDate": {
                    "type": "string",
                    "format": "date"
                },
                "status": RunStatus.json_schema(),
                "errorMessage": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "scheduleState": {
                    "oneOf": [
                        {"type": "null"},
                        ScheduleState.json_schema(),
                    ]
                },
                "statistics": {
                    "oneOf": [
                        {"type": "null"},
                        FeatureStoreExecutionStatistics.json_schema(),
                    ]
                }
            },
            "required": [
                "id",
                "created",
                "featureStoreId",
                "featureStoreVersionId",
                "commitId",
                "runStartDate",
                "runEndDate",
                "status",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreRun:
        """Validate and parse JSON data into an instance of FeatureStoreRun.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreRun.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreRun(
                id=FeatureStoreRunId.from_json(data["id"]),
                created=isodate.parse_datetime(data["created"]),
                featureStoreId=FeatureStoreId.from_json(data["featureStoreId"]),
                featureStoreVersionId=FeatureStoreVersionId.from_json(data["featureStoreVersionId"]),
                commitId=CommitId.from_json(data["commitId"]),
                runStartDate=datetime.date.fromisoformat(data["runStartDate"]),
                runEndDate=datetime.date.fromisoformat(data["runEndDate"]),
                status=RunStatus.from_json(data["status"]),
                errorMessage=(lambda v: v and str(v))(data.get("errorMessage", None)),
                scheduleState=(
                    lambda v: v and ScheduleState.from_json(v)
                )(
                    data.get("scheduleState", None)
                ),
                statistics=(
                    lambda v: v and FeatureStoreExecutionStatistics.from_json(v)
                )(
                    data.get("statistics", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureStoreRun",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "featureStoreId": self.featureStoreId.to_json(),
            "featureStoreVersionId": self.featureStoreVersionId.to_json(),
            "commitId": self.commitId.to_json(),
            "runStartDate": self.runStartDate.isoformat(),
            "runEndDate": self.runEndDate.isoformat(),
            "status": self.status.to_json(),
            "errorMessage": (lambda v: v and str(v))(self.errorMessage),
            "scheduleState": (lambda v: v and v.to_json())(self.scheduleState),
            "statistics": (lambda v: v and v.to_json())(self.statistics)
        }


@dataclasses.dataclass(frozen=True)
class FeatureStoreExecutionStatistics:
    """Statistics calculated during a feature store run.
    
    Args:
        base (ExecutionStatistics): A data field.
        rowCount (typing.Optional[int]): A data field.
        featureStatistics (typing.List[SummaryStatistics]): A data field.
        jobMetrics (typing.Optional[JobMetrics]): A data field.
    """
    
    base: ExecutionStatistics
    rowCount: typing.Optional[int]
    featureStatistics: typing.List[SummaryStatistics]
    jobMetrics: typing.Optional[JobMetrics]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreExecutionStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "base": ExecutionStatistics.json_schema(),
                "rowCount": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "integer"},
                    ]
                },
                "featureStatistics": {
                    "type": "array",
                    "item": SummaryStatistics.json_schema()
                },
                "jobMetrics": {
                    "oneOf": [
                        {"type": "null"},
                        JobMetrics.json_schema(),
                    ]
                }
            },
            "required": [
                "base",
                "featureStatistics",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreExecutionStatistics:
        """Validate and parse JSON data into an instance of FeatureStoreExecutionStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreExecutionStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreExecutionStatistics(
                base=ExecutionStatistics.from_json(data["base"]),
                rowCount=(lambda v: v and int(v))(data.get("rowCount", None)),
                featureStatistics=[SummaryStatistics.from_json(v) for v in data["featureStatistics"]],
                jobMetrics=(
                    lambda v: v and JobMetrics.from_json(v)
                )(
                    data.get("jobMetrics", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureStoreExecutionStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "base": self.base.to_json(),
            "rowCount": (lambda v: v and v)(self.rowCount),
            "featureStatistics": [v.to_json() for v in self.featureStatistics],
            "jobMetrics": (lambda v: v and v.to_json())(self.jobMetrics)
        }


@dataclasses.dataclass(frozen=True)
class ExecutionStatistics:
    """Statistics about a feature store run itself.
    
    Args:
        executionStartTime (datetime.datetime): A data field.
        executionEndTime (typing.Optional[datetime.datetime]): A data field.
    """
    
    executionStartTime: datetime.datetime
    executionEndTime: typing.Optional[datetime.datetime]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ExecutionStatistics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "executionStartTime": {
                    "type": "string",
                    "format": "date-time"
                },
                "executionEndTime": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                }
            },
            "required": [
                "executionStartTime",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ExecutionStatistics:
        """Validate and parse JSON data into an instance of ExecutionStatistics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ExecutionStatistics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ExecutionStatistics(
                executionStartTime=isodate.parse_datetime(data["executionStartTime"]),
                executionEndTime=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("executionEndTime", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ExecutionStatistics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "executionStartTime": self.executionStartTime.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "executionEndTime": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.executionEndTime)
        }


@dataclasses.dataclass(frozen=True)
class FeatureCompletedRunsDays:
    """Number of completed runs in days for a feature.
    
    Args:
        featureId (FeatureId): A data field.
        runsDaysCompleted (int): A data field.
    """
    
    featureId: FeatureId
    runsDaysCompleted: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureCompletedRunsDays data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "featureId": FeatureId.json_schema(),
                "runsDaysCompleted": {
                    "type": "integer"
                }
            },
            "required": [
                "featureId",
                "runsDaysCompleted",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureCompletedRunsDays:
        """Validate and parse JSON data into an instance of FeatureCompletedRunsDays.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureCompletedRunsDays.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureCompletedRunsDays(
                featureId=FeatureId.from_json(data["featureId"]),
                runsDaysCompleted=int(data["runsDaysCompleted"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureCompletedRunsDays",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "featureId": self.featureId.to_json(),
            "runsDaysCompleted": int(self.runsDaysCompleted)
        }



"""Generated implementation of anaml_object."""

# WARNING DO NOT EDIT
# This code was generated from anaml-object.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..entity import EntityId
from ..entity_mapping import EntityMappingId
from ..feature_id import FeatureId
from ..feature_set import FeatureSetId
from ..feature_store import FeatureStoreId
from ..feature_template import TemplateId
from ..table import TableId


@dataclasses.dataclass(frozen=True)
class AnamlObject(abc.ABC):
    """References to Anaml 'content' object."""
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> AnamlObject:
        """JSON schema for variant AnamlObject.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlObject:
        """Validate and parse JSON data into an instance of AnamlObject.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlObject.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing AnamlObject", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class AnamlEntity(AnamlObject):
    """A reference to an Entity object in Anaml.
    
    Args:
        id (EntityId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlentity"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: EntityId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlEntity data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": EntityId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlEntity:
        """Validate and parse JSON data into an instance of AnamlEntity.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlEntity.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlEntity(
                id=EntityId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlEntity",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlEntityMapping(AnamlObject):
    """A reference to an Entity Mapping object in Anaml.
    
    Args:
        id (EntityMappingId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlentitymapping"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: EntityMappingId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlEntityMapping data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": EntityMappingId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlEntityMapping:
        """Validate and parse JSON data into an instance of AnamlEntityMapping.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlEntityMapping.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlEntityMapping(
                id=EntityMappingId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlEntityMapping",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlFeature(AnamlObject):
    """A reference to an Feature object in Anaml.
    
    Args:
        id (FeatureId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlfeature"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: FeatureId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlFeature data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": FeatureId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlFeature:
        """Validate and parse JSON data into an instance of AnamlFeature.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlFeature.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlFeature(
                id=FeatureId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlFeature",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlFeatureTemplate(AnamlObject):
    """A reference to an Feature Template object in Anaml.
    
    Args:
        id (TemplateId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlfeaturetemplate"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: TemplateId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlFeatureTemplate data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": TemplateId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlFeatureTemplate:
        """Validate and parse JSON data into an instance of AnamlFeatureTemplate.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlFeatureTemplate.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlFeatureTemplate(
                id=TemplateId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlFeatureTemplate",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlFeatureSet(AnamlObject):
    """A reference to an Feature Set object in Anaml.
    
    Args:
        id (FeatureSetId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlfeatureset"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: FeatureSetId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlFeatureSet data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": FeatureSetId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlFeatureSet:
        """Validate and parse JSON data into an instance of AnamlFeatureSet.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlFeatureSet.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlFeatureSet(
                id=FeatureSetId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlFeatureSet",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlFeatureStore(AnamlObject):
    """A reference to an Feature Store object in Anaml.
    
    Args:
        id (FeatureStoreId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamlfeaturestore"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: FeatureStoreId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlFeatureStore data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": FeatureStoreId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlFeatureStore:
        """Validate and parse JSON data into an instance of AnamlFeatureStore.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlFeatureStore.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlFeatureStore(
                id=FeatureStoreId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlFeatureStore",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }


@dataclasses.dataclass(frozen=True)
class AnamlTable(AnamlObject):
    """A reference to an Table object in Anaml.
    
    Args:
        id (TableId): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "anamltable"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    id: TableId
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AnamlTable data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "id": TableId.json_schema()
            },
            "required": [
                "adt_type",
                "id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AnamlTable:
        """Validate and parse JSON data into an instance of AnamlTable.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AnamlTable.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AnamlTable(
                id=TableId.from_json(data["id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AnamlTable",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "id": self.id.to_json()
        }



"""Generated implementation of job_metrics."""

# WARNING DO NOT EDIT
# This code was generated from job-metrics.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401


@dataclasses.dataclass(frozen=True)
class JobMetrics:
    """Execution statistics for a job run.
    
    Args:
        appId (str): A data field.
        anamlVersion (str): A data field.
        executorCores (typing.Optional[str]): A data field.
        executorMemory (typing.Optional[str]): A data field.
        executorsMax (int): A data field.
        coresMax (int): A data field.
        runTimeSeconds (int): A data field.
        cpuTimeSeconds (int): A data field.
        gcTimeSeconds (int): A data field.
        bytesRead (int): A data field.
        bytesWritten (int): A data field.
        totalTasks (int): A data field.
        failedTasks (int): A data field.
        shuffleLocalBytesRead (int): A data field.
        shuffleRemoteBytesRead (int): A data field.
        shuffleBytesWritten (int): A data field.
    """
    
    appId: str
    anamlVersion: str
    executorCores: typing.Optional[str]
    executorMemory: typing.Optional[str]
    executorsMax: int
    coresMax: int
    runTimeSeconds: int
    cpuTimeSeconds: int
    gcTimeSeconds: int
    bytesRead: int
    bytesWritten: int
    totalTasks: int
    failedTasks: int
    shuffleLocalBytesRead: int
    shuffleRemoteBytesRead: int
    shuffleBytesWritten: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for JobMetrics data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "appId": {
                    "type": "string"
                },
                "anamlVersion": {
                    "type": "string"
                },
                "executorCores": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "executorMemory": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "executorsMax": {
                    "type": "integer"
                },
                "coresMax": {
                    "type": "integer"
                },
                "runTimeSeconds": {
                    "type": "integer"
                },
                "cpuTimeSeconds": {
                    "type": "integer"
                },
                "gcTimeSeconds": {
                    "type": "integer"
                },
                "bytesRead": {
                    "type": "integer"
                },
                "bytesWritten": {
                    "type": "integer"
                },
                "totalTasks": {
                    "type": "integer"
                },
                "failedTasks": {
                    "type": "integer"
                },
                "shuffleLocalBytesRead": {
                    "type": "integer"
                },
                "shuffleRemoteBytesRead": {
                    "type": "integer"
                },
                "shuffleBytesWritten": {
                    "type": "integer"
                }
            },
            "required": [
                "appId",
                "anamlVersion",
                "executorsMax",
                "coresMax",
                "runTimeSeconds",
                "cpuTimeSeconds",
                "gcTimeSeconds",
                "bytesRead",
                "bytesWritten",
                "totalTasks",
                "failedTasks",
                "shuffleLocalBytesRead",
                "shuffleRemoteBytesRead",
                "shuffleBytesWritten",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> JobMetrics:
        """Validate and parse JSON data into an instance of JobMetrics.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of JobMetrics.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return JobMetrics(
                appId=str(data["appId"]),
                anamlVersion=str(data["anamlVersion"]),
                executorCores=(lambda v: v and str(v))(data.get("executorCores", None)),
                executorMemory=(lambda v: v and str(v))(data.get("executorMemory", None)),
                executorsMax=int(data["executorsMax"]),
                coresMax=int(data["coresMax"]),
                runTimeSeconds=int(data["runTimeSeconds"]),
                cpuTimeSeconds=int(data["cpuTimeSeconds"]),
                gcTimeSeconds=int(data["gcTimeSeconds"]),
                bytesRead=int(data["bytesRead"]),
                bytesWritten=int(data["bytesWritten"]),
                totalTasks=int(data["totalTasks"]),
                failedTasks=int(data["failedTasks"]),
                shuffleLocalBytesRead=int(data["shuffleLocalBytesRead"]),
                shuffleRemoteBytesRead=int(data["shuffleRemoteBytesRead"]),
                shuffleBytesWritten=int(data["shuffleBytesWritten"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing JobMetrics",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "appId": str(self.appId),
            "anamlVersion": str(self.anamlVersion),
            "executorCores": (lambda v: v and str(v))(self.executorCores),
            "executorMemory": (lambda v: v and str(v))(self.executorMemory),
            "executorsMax": int(self.executorsMax),
            "coresMax": int(self.coresMax),
            "runTimeSeconds": self.runTimeSeconds,
            "cpuTimeSeconds": self.cpuTimeSeconds,
            "gcTimeSeconds": self.gcTimeSeconds,
            "bytesRead": self.bytesRead,
            "bytesWritten": self.bytesWritten,
            "totalTasks": int(self.totalTasks),
            "failedTasks": int(self.failedTasks),
            "shuffleLocalBytesRead": self.shuffleLocalBytesRead,
            "shuffleRemoteBytesRead": self.shuffleRemoteBytesRead,
            "shuffleBytesWritten": self.shuffleBytesWritten
        }



"""Generated implementation of restrictions."""

# WARNING DO NOT EDIT
# This code was generated from restrictions.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401


@dataclasses.dataclass(frozen=True)
class AttributeKey:
    """Attribute name.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AttributeKey data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AttributeKey:
        """Validate and parse JSON data into an instance of AttributeKey.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeKey.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AttributeKey(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing AttributeKey", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> AttributeKey:
        """Parse a JSON string such as a dictionary key."""
        return AttributeKey(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class AttributeDisplay:
    """Display details for an attribute value.
    
    Args:
        emoji (str): A data field.
        colour (str): A data field.
    """
    
    emoji: str
    colour: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AttributeDisplay data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "emoji": {
                    "type": "string"
                },
                "colour": {
                    "type": "string"
                }
            },
            "required": [
                "emoji",
                "colour",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AttributeDisplay:
        """Validate and parse JSON data into an instance of AttributeDisplay.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeDisplay.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AttributeDisplay(
                emoji=str(data["emoji"]),
                colour=str(data["colour"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AttributeDisplay",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "emoji": str(self.emoji),
            "colour": str(self.colour)
        }


@dataclasses.dataclass(frozen=True)
class AttributeChoice:
    """An attribute value.
    
    Args:
        value (str): A data field.
        display (typing.Optional[AttributeDisplay]): A data field.
    """
    
    value: str
    display: typing.Optional[AttributeDisplay]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AttributeChoice data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "value": {
                    "type": "string"
                },
                "display": {
                    "oneOf": [
                        {"type": "null"},
                        AttributeDisplay.json_schema(),
                    ]
                }
            },
            "required": [
                "value",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AttributeChoice:
        """Validate and parse JSON data into an instance of AttributeChoice.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeChoice.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AttributeChoice(
                value=str(data["value"]),
                display=(
                    lambda v: v and AttributeDisplay.from_json(v)
                )(
                    data.get("display", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AttributeChoice",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "value": str(self.value),
            "display": (lambda v: v and v.to_json())(self.display)
        }


class AttributeTarget(enum.Enum):
    """Object types an attribute may be applied to."""
    Cluster = "cluster"
    """Cluster attribute."""
    Destination = "destination"
    """Destination attribute."""
    Entity = "entity"
    """Entity attribute."""
    Feature = "feature"
    """Feature attribute."""
    FeatureSet = "featureset"
    """FeatureSet attribute."""
    FeatureStore = "featurestore"
    """FeatureStore attribute."""
    Source = "source"
    """Source attribute."""
    Table = "table"
    """Table attribute."""
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'AttributeTarget'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "cluster",
                        "destination",
                        "entity",
                        "feature",
                        "featureset",
                        "featurestore",
                        "source",
                        "table",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AttributeTarget:
        """Validate and parse JSON data into an instance of AttributeTarget.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeTarget.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AttributeTarget(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing AttributeTarget", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> AttributeTarget:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeTarget.
        """
        return AttributeTarget(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class AttributeRestriction(abc.ABC):
    """Defines attributes, values, and the objects they can be applied to.
    
    Args:
        appliesTo (typing.List[AttributeTarget]): A data field.
        description (str): A data field.
        key (AttributeKey): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    appliesTo: typing.List[AttributeTarget]
    description: str
    key: AttributeKey
    
    @classmethod
    def json_schema(cls) -> AttributeRestriction:
        """JSON schema for variant AttributeRestriction.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AttributeRestriction:
        """Validate and parse JSON data into an instance of AttributeRestriction.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AttributeRestriction.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing AttributeRestriction", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class EnumAttribute(AttributeRestriction):
    """Enumeration attribute with fixed choices.
    
    Args:
        key (AttributeKey): A data field.
        description (str): A data field.
        choices (typing.List[AttributeChoice]): A data field.
        appliesTo (typing.List[AttributeTarget]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "enumattribute"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    key: AttributeKey
    description: str
    choices: typing.List[AttributeChoice]
    appliesTo: typing.List[AttributeTarget]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for EnumAttribute data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "key": AttributeKey.json_schema(),
                "description": {
                    "type": "string"
                },
                "choices": {
                    "type": "array",
                    "item": AttributeChoice.json_schema()
                },
                "appliesTo": {
                    "type": "array",
                    "item": AttributeTarget.json_schema()
                }
            },
            "required": [
                "adt_type",
                "key",
                "description",
                "choices",
                "appliesTo",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> EnumAttribute:
        """Validate and parse JSON data into an instance of EnumAttribute.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of EnumAttribute.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return EnumAttribute(
                key=AttributeKey.from_json(data["key"]),
                description=str(data["description"]),
                choices=[AttributeChoice.from_json(v) for v in data["choices"]],
                appliesTo=[AttributeTarget.from_json(v) for v in data["appliesTo"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing EnumAttribute",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "key": self.key.to_json(),
            "description": str(self.description),
            "choices": [v.to_json() for v in self.choices],
            "appliesTo": [v.to_json() for v in self.appliesTo]
        }


@dataclasses.dataclass(frozen=True)
class FreeTextAttribute(AttributeRestriction):
    """Free text response attribute.
    
    Args:
        key (AttributeKey): A data field.
        description (str): A data field.
        appliesTo (typing.List[AttributeTarget]): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "freetextattribute"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    key: AttributeKey
    description: str
    appliesTo: typing.List[AttributeTarget]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FreeTextAttribute data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "key": AttributeKey.json_schema(),
                "description": {
                    "type": "string"
                },
                "appliesTo": {
                    "type": "array",
                    "item": AttributeTarget.json_schema()
                }
            },
            "required": [
                "adt_type",
                "key",
                "description",
                "appliesTo",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FreeTextAttribute:
        """Validate and parse JSON data into an instance of FreeTextAttribute.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FreeTextAttribute.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FreeTextAttribute(
                key=AttributeKey.from_json(data["key"]),
                description=str(data["description"]),
                appliesTo=[AttributeTarget.from_json(v) for v in data["appliesTo"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FreeTextAttribute",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "key": self.key.to_json(),
            "description": str(self.description),
            "appliesTo": [v.to_json() for v in self.appliesTo]
        }


@dataclasses.dataclass(frozen=True)
class AllowedAttributesResponse:
    """Policy for attributes permitted by the server configuration.
    
    Args:
        allowedAttributes (typing.Optional[typing.List[AttributeRestriction]]): A data field.
    """
    
    allowedAttributes: typing.Optional[typing.List[AttributeRestriction]]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AllowedAttributesResponse data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "allowedAttributes": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": AttributeRestriction.json_schema()},
                    ]
                }
            },
            "required": []
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AllowedAttributesResponse:
        """Validate and parse JSON data into an instance of AllowedAttributesResponse.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AllowedAttributesResponse.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AllowedAttributesResponse(
                allowedAttributes=(
                    lambda v: v and [AttributeRestriction.from_json(v) for v in v]
                )(
                    data.get("allowedAttributes", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AllowedAttributesResponse",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "allowedAttributes": (lambda v: v and [v.to_json() for v in v])(self.allowedAttributes)
        }


@dataclasses.dataclass(frozen=True)
class LabelRestriction:
    """Label restriction rule.
    
    Args:
        text (str): A data field.
        emoji (typing.Optional[str]): A data field.
        colour (typing.Optional[str]): A data field.
    """
    
    text: str
    emoji: typing.Optional[str]
    colour: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for LabelRestriction data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "text": {
                    "type": "string"
                },
                "emoji": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "colour": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "text",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> LabelRestriction:
        """Validate and parse JSON data into an instance of LabelRestriction.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of LabelRestriction.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return LabelRestriction(
                text=str(data["text"]),
                emoji=(lambda v: v and str(v))(data.get("emoji", None)),
                colour=(lambda v: v and str(v))(data.get("colour", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing LabelRestriction",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "text": str(self.text),
            "emoji": (lambda v: v and str(v))(self.emoji),
            "colour": (lambda v: v and str(v))(self.colour)
        }


@dataclasses.dataclass(frozen=True)
class AllowedLabelsResponse:
    """Policy for labels permitted by the server configuration.
    
    Args:
        allowedLabels (typing.Optional[typing.List[LabelRestriction]]): A data field.
    """
    
    allowedLabels: typing.Optional[typing.List[LabelRestriction]]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AllowedLabelsResponse data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "allowedLabels": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": LabelRestriction.json_schema()},
                    ]
                }
            },
            "required": []
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AllowedLabelsResponse:
        """Validate and parse JSON data into an instance of AllowedLabelsResponse.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AllowedLabelsResponse.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AllowedLabelsResponse(
                allowedLabels=(
                    lambda v: v and [LabelRestriction.from_json(v) for v in v]
                )(
                    data.get("allowedLabels", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AllowedLabelsResponse",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "allowedLabels": (lambda v: v and [v.to_json() for v in v])(self.allowedLabels)
        }



"""Generated implementation of feature_run_summaries."""

# WARNING DO NOT EDIT
# This code was generated from feature-run-summaries.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..feature_store import FeatureStoreId
from ..jobs import FeatureStoreRunId
from ..summary_statistics import SummaryStatistics


@dataclasses.dataclass(frozen=True)
class FeatureRunSummary:
    """Data describing a feature store run and summarising the generated features.
    
    Args:
        featureStore (FeatureStoreId): A data field.
        featureRun (FeatureStoreRunId): A data field.
        runStartDate (datetime.date): A data field.
        runEndDate (datetime.date): A data field.
        runTime (datetime.datetime): A data field.
        stats (SummaryStatistics): A data field.
    """
    
    featureStore: FeatureStoreId
    featureRun: FeatureStoreRunId
    runStartDate: datetime.date
    runEndDate: datetime.date
    runTime: datetime.datetime
    stats: SummaryStatistics
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureRunSummary data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "featureStore": FeatureStoreId.json_schema(),
                "featureRun": FeatureStoreRunId.json_schema(),
                "runStartDate": {
                    "type": "string",
                    "format": "date"
                },
                "runEndDate": {
                    "type": "string",
                    "format": "date"
                },
                "runTime": {
                    "type": "string",
                    "format": "date-time"
                },
                "stats": SummaryStatistics.json_schema()
            },
            "required": [
                "featureStore",
                "featureRun",
                "runStartDate",
                "runEndDate",
                "runTime",
                "stats",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureRunSummary:
        """Validate and parse JSON data into an instance of FeatureRunSummary.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureRunSummary.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureRunSummary(
                featureStore=FeatureStoreId.from_json(data["featureStore"]),
                featureRun=FeatureStoreRunId.from_json(data["featureRun"]),
                runStartDate=datetime.date.fromisoformat(data["runStartDate"]),
                runEndDate=datetime.date.fromisoformat(data["runEndDate"]),
                runTime=isodate.parse_datetime(data["runTime"]),
                stats=SummaryStatistics.from_json(data["stats"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FeatureRunSummary",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "featureStore": self.featureStore.to_json(),
            "featureRun": self.featureRun.to_json(),
            "runStartDate": self.runStartDate.isoformat(),
            "runEndDate": self.runEndDate.isoformat(),
            "runTime": self.runTime.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "stats": self.stats.to_json()
        }



"""Generated implementation of health."""

# WARNING DO NOT EDIT
# This code was generated from health.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401


@dataclasses.dataclass(frozen=True)
class HealthStatus(abc.ABC):
    """Health status types.
    
    Args:
        indicator (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    indicator: str
    
    @classmethod
    def json_schema(cls) -> HealthStatus:
        """JSON schema for variant HealthStatus.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> HealthStatus:
        """Validate and parse JSON data into an instance of HealthStatus.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of HealthStatus.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing HealthStatus", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class Healthy(HealthStatus):
    """Healthy status.
    
    Args:
        indicator (str): A data field.
        message (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "healthy"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    indicator: str
    message: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Healthy data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "indicator": {
                    "type": "string"
                },
                "message": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "indicator",
                "message",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Healthy:
        """Validate and parse JSON data into an instance of Healthy.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Healthy.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Healthy(
                indicator=str(data["indicator"]),
                message=str(data["message"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Healthy",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "indicator": str(self.indicator),
            "message": str(self.message)
        }


@dataclasses.dataclass(frozen=True)
class Unhealthy(HealthStatus):
    """Unhealthy status.
    
    Args:
        indicator (str): A data field.
        reason (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "unhealthy"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    indicator: str
    reason: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Unhealthy data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "indicator": {
                    "type": "string"
                },
                "reason": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "indicator",
                "reason",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Unhealthy:
        """Validate and parse JSON data into an instance of Unhealthy.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Unhealthy.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Unhealthy(
                indicator=str(data["indicator"]),
                reason=str(data["reason"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Unhealthy",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "indicator": str(self.indicator),
            "reason": str(self.reason)
        }


@dataclasses.dataclass(frozen=True)
class HealthResponse:
    """Health status across all health indicators.
    
    Args:
        healthy (bool): A data field.
        indicators (typing.List[HealthStatus]): A data field.
    """
    
    healthy: bool
    indicators: typing.List[HealthStatus]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for HealthResponse data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "healthy": {
                    "type": "boolean"
                },
                "indicators": {
                    "type": "array",
                    "item": HealthStatus.json_schema()
                }
            },
            "required": [
                "healthy",
                "indicators",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> HealthResponse:
        """Validate and parse JSON data into an instance of HealthResponse.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of HealthResponse.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return HealthResponse(
                healthy=bool(data["healthy"]),
                indicators=[HealthStatus.from_json(v) for v in data["indicators"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing HealthResponse",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "healthy": self.healthy,
            "indicators": [v.to_json() for v in self.indicators]
        }



"""Generated implementation of checks."""

# WARNING DO NOT EDIT
# This code was generated from checks.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401

from ..commit import CommitId
from ..user import UserId


class CheckStatus(enum.Enum):
    """Status of an external check run."""
    Pending = "pending"
    Running = "running"
    Completed = "completed"
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'CheckStatus'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "pending",
                        "running",
                        "completed",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CheckStatus:
        """Validate and parse JSON data into an instance of CheckStatus.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CheckStatus.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CheckStatus(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing CheckStatus", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> CheckStatus:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of CheckStatus.
        """
        return CheckStatus(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


class CheckConclusion(enum.Enum):
    """Outcome of an external check run."""
    Cancelled = "cancelled"
    Failure = "failure"
    Success = "success"
    Skipped = "skipped"
    TimedOut = "timedout"
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'CheckConclusion'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "cancelled",
                        "failure",
                        "success",
                        "skipped",
                        "timedout",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CheckConclusion:
        """Validate and parse JSON data into an instance of CheckConclusion.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CheckConclusion.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CheckConclusion(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing CheckConclusion", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> CheckConclusion:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of CheckConclusion.
        """
        return CheckConclusion(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class CheckId:
    """Unique identifier of an external check.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CheckId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CheckId:
        """Validate and parse JSON data into an instance of CheckId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CheckId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CheckId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing CheckId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> CheckId:
        """Parse a JSON string such as a dictionary key."""
        return CheckId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class CheckComponent:
    """Results for a specific component in an external check.
    
    Args:
        key (str): A data field.
        value (str): A data field.
    """
    
    key: str
    value: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CheckComponent data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "key": {
                    "type": "string"
                },
                "value": {
                    "type": "string"
                }
            },
            "required": [
                "key",
                "value",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CheckComponent:
        """Validate and parse JSON data into an instance of CheckComponent.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CheckComponent.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CheckComponent(
                key=str(data["key"]),
                value=str(data["value"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CheckComponent",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "key": str(self.key),
            "value": str(self.value)
        }


@dataclasses.dataclass(frozen=True)
class Check:
    """Details of an external check.
    
    Args:
        id (CheckId): A data field.
        name (str): A data field.
        summary (typing.Optional[str]): A data field.
        commit (CommitId): A data field.
        created_by (UserId): A data field.
        started (typing.Optional[datetime.datetime]): A data field.
        completed (typing.Optional[datetime.datetime]): A data field.
        status (CheckStatus): A data field.
        conclusion (typing.Optional[CheckConclusion]): A data field.
        components (typing.List[CheckComponent]): A data field.
        details_url (typing.Optional[str]): A data field.
    """
    
    id: CheckId
    name: str
    summary: typing.Optional[str]
    commit: CommitId
    created_by: UserId
    started: typing.Optional[datetime.datetime]
    completed: typing.Optional[datetime.datetime]
    status: CheckStatus
    conclusion: typing.Optional[CheckConclusion]
    components: typing.List[CheckComponent]
    details_url: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Check data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": CheckId.json_schema(),
                "name": {
                    "type": "string"
                },
                "summary": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "commit": CommitId.json_schema(),
                "created_by": UserId.json_schema(),
                "started": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "completed": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "status": CheckStatus.json_schema(),
                "conclusion": {
                    "oneOf": [
                        {"type": "null"},
                        CheckConclusion.json_schema(),
                    ]
                },
                "components": {
                    "type": "array",
                    "item": CheckComponent.json_schema()
                },
                "details_url": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "id",
                "name",
                "commit",
                "created_by",
                "status",
                "components",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Check:
        """Validate and parse JSON data into an instance of Check.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Check.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Check(
                id=CheckId.from_json(data["id"]),
                name=str(data["name"]),
                summary=(lambda v: v and str(v))(data.get("summary", None)),
                commit=CommitId.from_json(data["commit"]),
                created_by=UserId.from_json(data["created_by"]),
                started=(lambda v: v and isodate.parse_datetime(v))(data.get("started", None)),
                completed=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("completed", None)
                ),
                status=CheckStatus.from_json(data["status"]),
                conclusion=(
                    lambda v: v and CheckConclusion.from_json(v)
                )(
                    data.get("conclusion", None)
                ),
                components=[CheckComponent.from_json(v) for v in data["components"]],
                details_url=(lambda v: v and str(v))(data.get("details_url", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Check",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "name": str(self.name),
            "summary": (lambda v: v and str(v))(self.summary),
            "commit": self.commit.to_json(),
            "created_by": self.created_by.to_json(),
            "started": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.started),
            "completed": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.completed),
            "status": self.status.to_json(),
            "conclusion": (lambda v: v and v.to_json())(self.conclusion),
            "components": [v.to_json() for v in self.components],
            "details_url": (lambda v: v and str(v))(self.details_url)
        }


@dataclasses.dataclass(frozen=True)
class CheckCreationRequest:
    """Request to create a new external check.
    
    Args:
        name (str): A data field.
        summary (typing.Optional[str]): A data field.
        started (typing.Optional[datetime.datetime]): A data field.
        completed (typing.Optional[datetime.datetime]): A data field.
        status (CheckStatus): A data field.
        conclusion (typing.Optional[CheckConclusion]): A data field.
        components (typing.Optional[typing.List[CheckComponent]]): A data field.
        details_url (typing.Optional[str]): A data field.
    """
    
    name: str
    summary: typing.Optional[str]
    started: typing.Optional[datetime.datetime]
    completed: typing.Optional[datetime.datetime]
    status: CheckStatus
    conclusion: typing.Optional[CheckConclusion]
    components: typing.Optional[typing.List[CheckComponent]]
    details_url: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CheckCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": {
                    "type": "string"
                },
                "summary": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "started": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "completed": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string", "format": "date-time"},
                    ]
                },
                "status": CheckStatus.json_schema(),
                "conclusion": {
                    "oneOf": [
                        {"type": "null"},
                        CheckConclusion.json_schema(),
                    ]
                },
                "components": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "array", "item": CheckComponent.json_schema()},
                    ]
                },
                "details_url": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "name",
                "status",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CheckCreationRequest:
        """Validate and parse JSON data into an instance of CheckCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CheckCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CheckCreationRequest(
                name=str(data["name"]),
                summary=(lambda v: v and str(v))(data.get("summary", None)),
                started=(lambda v: v and isodate.parse_datetime(v))(data.get("started", None)),
                completed=(
                    lambda v: v and isodate.parse_datetime(v)
                )(
                    data.get("completed", None)
                ),
                status=CheckStatus.from_json(data["status"]),
                conclusion=(
                    lambda v: v and CheckConclusion.from_json(v)
                )(
                    data.get("conclusion", None)
                ),
                components=(
                    lambda v: v and [CheckComponent.from_json(v) for v in v]
                )(
                    data.get("components", None)
                ),
                details_url=(lambda v: v and str(v))(data.get("details_url", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CheckCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": str(self.name),
            "summary": (lambda v: v and str(v))(self.summary),
            "started": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.started),
            "completed": (lambda v: v and v.strftime('%Y-%m-%dT%H:%M:%S.%f%z'))(self.completed),
            "status": self.status.to_json(),
            "conclusion": (lambda v: v and v.to_json())(self.conclusion),
            "components": (lambda v: v and [v.to_json() for v in v])(self.components),
            "details_url": (lambda v: v and str(v))(self.details_url)
        }
